from pexpect import spawn, TIMEOUT, EOF

from .const import CONNECT_ERROR_EOF, CONNECT_ERROR_TIMEOUT, LOGIN_ERROR_LOGIN_INCORRECT, \
    CONNECT_ERROR_PRIVILEGED_ALREADY_LOGGED_IN
from .exceptions import AuthenticationError


class RuckusSSH(spawn):
    """Ruckus SSH client."""

    def __init__(self, timeout=30, maxread=2000, searchwindowsize=None,
                 logfile=None, cwd=None, env=None, ignore_sighup=True, echo=True,
                 encoding=None, codec_errors='strict', use_poll=False) -> None:

        spawn.__init__(self, None, timeout=timeout, maxread=maxread,
                       searchwindowsize=searchwindowsize, logfile=logfile,
                       cwd=cwd, env=env, ignore_sighup=ignore_sighup, echo=echo,
                       encoding=encoding, codec_errors=codec_errors, use_poll=use_poll)

    def login(self, host, username=None, password='', login_timeout=10) -> bool:
        """Takes the host, username, and password, and logs into the Ruckus device."""
        spawn._spawn(self, f"ssh {host}")

        login_regex_array = ["Please login: ", "(?i)are you sure you want to continue connecting", EOF, TIMEOUT]

        i = self.expect(login_regex_array, timeout=login_timeout)
        if i == 1:
            # New certificate -- always accept it.
            # This is what you get if SSH does not have the remote host's
            # public key stored in the 'known_hosts' cache.
            self.sendline("yes")
            i = self.expect(login_regex_array, timeout=login_timeout)
        if i == 2:
            raise ConnectionError(CONNECT_ERROR_EOF)
        if i == 3:
            raise ConnectionError(CONNECT_ERROR_TIMEOUT)

        self.sendline(username)

        self.expect("Password: ")
        self.sendline(password)

        i = self.expect(["> ", "Login incorrect"])
        if i == 1:
            raise AuthenticationError(LOGIN_ERROR_LOGIN_INCORRECT)

        return True

    def run_privileged(self, cmd: str) -> str:
        self.enable()
        self.sendline(cmd)
        self.expect("\n")
        self.prompt()
        result = self.before
        self.disable()
        return result

    def prompt(self, timeout=-1) -> int:
        """Wait for prompt and determine the current level of permissions."""
        if timeout == -1:
            timeout = self.timeout
        i = self.expect(
            ["ruckus> ", "ruckus# ", "A privileged user is already logged in", EOF, TIMEOUT],
            timeout=timeout
        )
        if i == 2:
            self.prompt(timeout)
            raise ConnectionError(CONNECT_ERROR_PRIVILEGED_ALREADY_LOGGED_IN)
        if i == 3:
            raise ConnectionError(CONNECT_ERROR_EOF)
        if i == 4:
            raise ConnectionError(CONNECT_ERROR_TIMEOUT)
        return i

    def enable(self, force=False) -> None:
        """Enable privileged commands"""
        cmd = "enable"
        if force:
            cmd += " force"

        self.sendline(cmd)
        self.prompt()

    def disable(self) -> None:
        self.sendline("disable")
        self.prompt()



"""Constants used in pyruckus."""
import re

# Error strings
CONNECT_ERROR_EOF = "Could not establish connection to host"
CONNECT_ERROR_TIMEOUT = "Timed out while waiting for client"
CONNECT_ERROR_PRIVILEGED_ALREADY_LOGGED_IN = "A privileged user is already logged in"
LOGIN_ERROR_LOGIN_INCORRECT = "Login incorrect"



"""Packaging logic for pyruckus."""

from setuptools import setup

with open("README.md", "r") as file:
    long_description = file.read()

setup(
    name="pyruckus",
    version="0.11",
    description="Python API to interact with a Ruckus Unleashed device.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gabe565/pyruckus",
    author="Gabe Cook",
    author_email="gabe565@gmail.com",
    license="MIT",
    install_requires=["pexpect>=4.0"],
    packages=["pyruckus"],
    zip_safe=True,
    python_requires=">=3.6",
)



from .RuckusSSH import RuckusSSH


class Ruckus:
    """Class for communicating with the device."""

    def __init__(self, host: str, username: str, password: str, login_timeout=15, timeout=10) -> None:
        """Set runtime configuration."""
        self.host = host
        self.username = username
        self.password = password
        self.login_timeout = login_timeout
        self.timeout = timeout

        self.ssh = None
        self.connect()

    def __del__(self) -> None:
        """Disconnect on delete."""
        self.disconnect()

    def connect(self) -> None:
        """Create SSH connection and login."""
        ssh = RuckusSSH(encoding="utf-8")
        ssh.login(self.host, username=self.username, password=self.password, login_timeout=self.login_timeout)
        self.ssh = ssh

    def disconnect(self) -> None:
        """Close the SSH session."""
        if self.ssh and self.ssh.isalive():
            self.ssh.close()

    def clients(self) -> dict:
        """Pull active clients from the device."""
        if not self.ssh.isalive():
            self.connect()

        result = self.ssh.run_privileged("show current-active-clients all")
        result, _, _ = result.partition("Last 300 Events/Activities:")

        result = self.__parse_kv(result)

        return {e['Mac Address']: e for e in result['Current Active Clients']['Clients']}

    def mesh_name(self) -> str:
        """Pull the current mesh name."""
        try:
            return self.mesh_info()['Mesh Settings']['Mesh Name(ESSID)']
        except KeyError:
            return 'Ruckus Mesh'

    @staticmethod
    def __parse_kv(response) -> dict:
        """Parse Ruckus nested key-value output into a dict."""
        root = {}
        indent = 0

        node = root
        breadcrumbs = [root]
        for line in response.splitlines():
            # Skip empty lines
            if not line.strip():
                continue

            # Line is a "header" instead of a key-value pair
            is_header = line.endswith(":") and "= " not in line

            prev_indent = indent
            indent = len(line) - len(line.lstrip())

            # If the indent has decreased, remove nodes from the breadcrumbs
            if indent < prev_indent:
                difference = int((indent - prev_indent) / 2)
                breadcrumbs = breadcrumbs[:difference]
                node = breadcrumbs[-1]

            if is_header:
                # Remove colon, then strip whitespace
                line = line[:-1].strip()
                parent_node = breadcrumbs[-1]
                node = {}

                # If current header already exists, convert to list
                if line in parent_node:
                    if isinstance(parent_node[line], list):
                        parent_node[line].append(node)
                    else:
                        prev_node = parent_node[line]
                        parent_node[line] = [prev_node, node]
                else:
                    parent_node[line] = node

                breadcrumbs.append(node)
            else:
                key, _, value = line.partition("= ")
                if key and value:
                    node[key.strip()] = value.strip()

        return root

    def mesh_info(self) -> dict:
        """Pull the current mesh name."""
        if not self.ssh.isalive():
            self.connect()

        result = self.ssh.run_privileged("show mesh info")

        return self.__parse_kv(result)

    def system_info(self) -> dict:
        """Pull the system info."""
        if not self.ssh.isalive():
            self.connect()

        result = self.ssh.run_privileged("show sysinfo")

        return self.__parse_kv(result)

    def ap_info(self) -> dict:
        """Pull info about current access points."""
        if not self.ssh.isalive():
            self.connect()

        result = self.ssh.run_privileged("show ap all")

        return self.__parse_kv(result)
